""" do file io """

# pylint: disable=invalid-name

from __future__ import annotations

import json
from math import ceil
from os.path import exists
from pathlib import Path
from re import findall
from shlex import split
from typing import Any, Iterable
from xml.etree import ElementTree
from xml.etree.ElementTree import Element, SubElement

import swatch
import tomlkit
import yaml
from colormath.color_objects import CMYKColor, ColorBase, LabColor, sRGBColor
from defusedxml.ElementTree import parse
from defusedxml.minidom import parseString
from metprint import FHFormatter, Logger, LogType
from PIL import Image, ImageDraw

from colourswatch.colourswatch import Colour, ColourSwatch
from colourswatch.GimpGplPalette import GimpGplPalette


def prettify(
	elem: Element, indent: str = "\t", doctype: str = '<?xml version="1.0" encoding="utf-8"?>'
):
	"""Return a pretty-printed XML string for the Element."""
	rough_string = ElementTree.tostring(elem, "utf-8")
	reparsed = parseString(rough_string)
	reparsed = reparsed.toprettyxml(indent=indent).split("\n")[1:]
	reparsed.insert(0, doctype)
	return "\n".join(reparsed)


def extNotRecognised(fileName: str):
	"""Output the file extension not recognised error"""
	exts = ', "'.join(
		[
			"gpl",
			"yaml",
			"colors",
			"spl",
			"skp",
			"soc",
			"txt",
			"acbl",
			"xml",
			"pal",
			"hpl",
			"toml",
			"json",
			"ase",
			"png",
			"jpg",
			"webp",
			"svg",
		]
	)
	Logger(FHFormatter()).logPrint(
		f'File extension is not recognised for file: {fileName}! Must be one of "{exts}"',
		LogType.ERROR,
	)


def openColourSwatch(file: str) -> ColourSwatch:
	"""Open a colour swatch file into a layer image object

	Args:
		file (str): path/ filename

	Raises:
		FileExistsError: [description]
		ValueError: [description]

	Returns:
		ColourSwatch: a colour swatch object
	"""
	functionMap = {
		"gpl": openSwatch_GPL,
		"yaml": openSwatch_YAML,
		"colors": openSwatch_COLOR,
		"spl": openSwatch_SPL,
		"skp": openSwatch_SKP,
		"soc": openSwatch_SOC,
		"txt": openSwatch_TXT,
		"acbl": openSwatch_ACBL,
		"xml": openSwatch_XML,
		"pal": openSwatch_CDPAL,
		"hpl": openSwatch_HPL,
		"toml": openSwatch_TOML,
		"json": openSwatch_JSON,
		"png": openSwatch_PNG,
		"jpg": openSwatch_IMAGE,
		"webp": openSwatch_IMAGE,
		"ase": openSwatch_ASE,
		"svg": openSwatch_SVG,
	}
	if not exists(file):
		Logger(FHFormatter()).logPrint(f"{file} does not exist", LogType.ERROR)
		raise FileExistsError
	fileExt = file.split(".")[-1].lower()
	if fileExt not in functionMap:
		extNotRecognised(file)
		raise ValueError
	return functionMap[fileExt](file)


def saveColourSwatch(fileName: str, colourSwatch: ColourSwatch) -> None:
	"""Save a colour swatch to a file

	Args:
		fileName (str): path/ filename
		colourSwatch (ColourSwatch): the colour swatch to save

	Raises:
		ValueError: [description]

	Returns:
		None: [description]
	"""
	functionMap = {
		"gpl": saveSwatch_GPL,
		"yaml": saveSwatch_YAML,
		"colors": saveSwatch_COLOR,
		"spl": saveSwatch_SPL,
		"skp": saveSwatch_SKP,
		"soc": saveSwatch_SOC,
		"txt": saveSwatch_TXT,
		"acbl": saveSwatch_ACBL,
		"xml": saveSwatch_XML,
		"pal": saveSwatch_CDPAL,
		"hpl": saveSwatch_HPL,
		"toml": saveSwatch_TOML,
		"json": saveSwatch_JSON,
		"png": saveSwatch_IMAGE,
		"jpg": saveSwatch_IMAGE,
		"webp": saveSwatch_IMAGE,
		"ase": saveSwatch_ASE,
		"svg": saveSwatch_SVG,
	}
	fileExt = fileName.split(".")[-1].lower()
	if fileExt not in functionMap:
		extNotRecognised(fileName)
		raise ValueError
	return functionMap[fileExt](fileName, colourSwatch)


def getColourFromLine(
	line: str,
	lineno: int,
	colourSpaceSize: int = 3,
	colourSpace: ColorBase = sRGBColor,
	divider: int = 255,
):
	"""getColourFromLine"""
	parts = line.split(None)
	return Colour(
		" ".join(parts[colourSpaceSize:]) if len(parts) > colourSpaceSize else f"colour{lineno}",
		colour=colourSpace(*[float(col) / divider for col in parts[:colourSpaceSize]]),
		nameNull=len(parts) <= colourSpaceSize,
	)


def getSwatchFromFileName(file: str, colours: list[Colour]):
	"""getSwatchFromFileName"""
	return ColourSwatch(file.replace("\\", "/").split("/")[-1].split(".")[0], colours)


def getWriteOutColour(
	colour: Iterable[Any], convertType: type = int, multiplier: int = 255
) -> list[Any]:
	"""getWriteOutColour"""
	return [convertType(col * multiplier) for col in colour]


### GPL ###
def openSwatch_GPL(file: str):
	"""Open a .GPL into a colour swatch"""
	project = GimpGplPalette(file)
	colours = []
	for index, colour in enumerate(project.colors):
		colours.append(
			Colour(
				project.colorNames[index]
				if project.colorNames[index] is not None
				else f"colour{index}",
				colour=sRGBColor(colour[0], colour[1], colour[2], True),
				nameNull=project.colorNames[index] is None,
			)
		)
	return ColourSwatch(project.name, colours=colours)


def saveSwatch_GPL(fileName: str, colourSwatch: ColourSwatch):
	"""Save a colour swatch as .GPL"""
	project = GimpGplPalette()
	project.name = colourSwatch.name
	project.colors = [
		getWriteOutColour(colour.toRGB().get_value_tuple()) for colour in colourSwatch.colours
	]
	project.colorNames = [
		colour.name if not colour.nameNull else None for colour in colourSwatch.colours
	]
	project.save(fileName)


### YAML ###
def openSwatch_YAML(file: str) -> ColourSwatch:
	"""Open a .YAML into a colour swatch"""
	project = yaml.safe_load(Path(file).read_text(encoding="utf-8"))
	swatchName = project.pop("scheme") if "scheme" in project else project.pop("name")
	swatchAuthor = project.pop("author") if "author" in project else None
	return ColourSwatch(
		swatchName,
		colours=[
			Colour(
				key,
				colour=sRGBColor(
					int(project[key][0:2], 16),
					int(project[key][2:4], 16),
					int(project[key][4:6], 16),
					True,
				),
			)
			for key in project
		],
		author=swatchAuthor,
	)


def saveSwatch_YAML(fileName: str, colourSwatch: ColourSwatch):
	"""Save a colour swatch as .YAML"""
	yamldict = {"scheme": colourSwatch.name, "author": colourSwatch.author}
	for colour in colourSwatch.colours:
		yamldict[colour.name] = "".join(colour.getRGB255Hex())
	Path(fileName).write_text(yaml.safe_dump(yamldict), encoding="utf-8")


### JSON ###
def openSwatch_JSON(file: str) -> ColourSwatch:
	"""Open a .JSON into a colour swatch"""
	project = json.loads(Path(file).read_text(encoding="utf-8"))
	swatchName = project.pop("scheme") if "scheme" in project else project.pop("name")
	swatchAuthor = project.pop("author") if "author" in project else None
	return ColourSwatch(
		swatchName,
		colours=[
			Colour(
				key,
				colour=sRGBColor(
					int(project[key][0:2], 16),
					int(project[key][2:4], 16),
					int(project[key][4:6], 16),
					True,
				),
			)
			for key in project
		],
		author=swatchAuthor,
	)


def saveSwatch_JSON(fileName: str, colourSwatch: ColourSwatch):
	"""Save a colour swatch as .JSON"""
	jsondict = {"scheme": colourSwatch.name, "author": colourSwatch.author}
	for colour in colourSwatch.colours:
		jsondict[colour.name] = "".join(colour.getRGB255Hex())
	Path(fileName).write_text(json.dumps(jsondict, indent="\t"), encoding="utf-8")


### TOML ###
def openSwatch_TOML(file: str) -> ColourSwatch:
	"""Open a .TOML into a colour swatch"""
	project = tomlkit.loads(Path(file).read_text(encoding="utf-8"))
	swatchName = project.pop("scheme") if "scheme" in project else project.pop("name")
	swatchAuthor = project.pop("author") if "author" in project else None
	return ColourSwatch(
		swatchName,
		colours=[
			Colour(
				key,
				colour=sRGBColor(
					int(project[key][0:2], 16),
					int(project[key][2:4], 16),
					int(project[key][4:6], 16),
					True,
				),
			)
			for key in project
		],
		author=swatchAuthor,
	)


def saveSwatch_TOML(fileName: str, colourSwatch: ColourSwatch):
	"""Save a colour swatch as .TOML"""
	tomldict = {"scheme": colourSwatch.name, "author": colourSwatch.author}
	for colour in colourSwatch.colours:
		tomldict[colour.name] = "".join(colour.getRGB255Hex())
	Path(fileName).write_text(tomlkit.dumps(tomldict), encoding="utf-8")


### COLOR ###
def openSwatch_COLOR(file: str) -> ColourSwatch:
	"""Open a .COLOR into a colour swatch"""
	colours = []
	for lineno, line in enumerate(Path(file).read_text(encoding="utf-8").splitlines(False)[1:]):
		colours.append(getColourFromLine(line, lineno))
	return getSwatchFromFileName(file, colours)


def saveSwatch_COLOR(fileName: str, colourSwatch: ColourSwatch):
	"""Save a colour swatch as .COLOR"""
	Path(fileName).write_text(
		"KDE RGB Palette\n"
		+ "\n".join(
			[
				" ".join([str(col) for col in colour.getRGB255()]) + f"\t{colour.name}"
				for colour in colourSwatch.colours
			]
		)
		+ "\n",
		encoding="utf-8",
	)


### SPL ###
def openSwatch_SPL(file: str) -> ColourSwatch:
	"""Open a .SPL into a colour swatch"""
	colours = []
	for lineno, line in enumerate(Path(file).read_text(encoding="utf-8").splitlines(False)[1:]):
		colours.append(getColourFromLine(line, lineno, divider=1))
	return getSwatchFromFileName(file, colours)


def saveSwatch_SPL(fileName: str, colourSwatch: ColourSwatch):
	"""Save a colour swatch as .SPL"""
	Path(fileName).write_text(
		"##Sketch RGBPalette 0\n"
		+ "\n".join(
			[
				" ".join(
					[
						f"{col:.6f}"
						for col in getWriteOutColour(colour.toRGB().get_value_tuple(), float, 1)
					]
				)
				+ f"\t{colour.name}"
				for colour in colourSwatch.colours
			]
		)
		+ "\n",
		encoding="utf-8",
	)


### SKP ###
def openSwatch_SKP(file: str) -> ColourSwatch:
	"""Open a .SKP into a colour swatch"""
	xmlrep = parse(file).getroot()
	cType = xmlrep[0].attrib["type"]
	colours = []
	for colour in xmlrep[1:]:
		cName = colour.attrib["name"]
		if cType == "CMYK":
			cColour = CMYKColor(
				float(colour.attrib["c"]),
				float(colour.attrib["m"]),
				float(colour.attrib["y"]),
				float(colour.attrib["k"]),
			)
		else:
			cColour = sRGBColor(
				float(colour.attrib["r"]), float(colour.attrib["g"]), float(colour.attrib["b"])
			)
		colours.append(Colour(cName, cColour))
	return ColourSwatch(xmlrep[0].attrib["name"], colours)


def saveSwatch_SKP(fileName: str, colourSwatch: ColourSwatch):
	"""Save a colour swatch as .SKP"""
	root = Element("palette")
	SubElement(root, "description", {"type": "CMYK", "name": colourSwatch.name})
	for colour in colourSwatch.colours:
		c, m, y, k = getWriteOutColour(colour.toCMYK().get_value_tuple(), float, 1)
		SubElement(
			root,
			"color",
			{"c": str(c), "m": str(m), "y": str(y), "k": str(k), "name": colour.name},
		)
	Path(fileName).write_text(prettify(root), encoding="utf-8")


### SOC ###
def openSwatch_SOC(file: str) -> ColourSwatch:
	"""Open a .SOC into a colour swatch"""
	xmlrep = parse(file).getroot()
	colours = []
	for colour in xmlrep:
		cName = colour.attrib["{http://openoffice.org/2000/drawing}name"]
		col = colour.attrib["{http://openoffice.org/2000/drawing}color"][1:]
		cColour = sRGBColor(int(col[0:2], 16), int(col[2:4], 16), int(col[4:6], 16), True)
		colours.append(Colour(cName, cColour))
	return getSwatchFromFileName(file, colours)


def saveSwatch_SOC(fileName: str, colourSwatch: ColourSwatch):
	"""Save a colour swatch as .SOC"""
	root = Element("office_color_table")
	for colour in colourSwatch.colours:
		SubElement(
			root,
			"draw_color",
			{
				"draw_name": colour.name,
				"draw_color": "#"
				+ "".join(
					[f"{col:02x}" for col in getWriteOutColour(colour.toRGB().get_value_tuple())]
				),
			},
		)
	Path(fileName).write_text(
		prettify(root, indent=" ", doctype='<?xml version="1.0" encoding="UTF-8"?>\n')
		.replace(
			"<office_color_table>",
			'<office:color-table xmlns:office="http://openoffice.org/2000/office" '
			+ 'xmlns:style="http://openoffice.org/2000/style" '
			+ 'xmlns:text="http://openoffice.org/2000/text" '
			+ 'xmlns:table="http://openoffice.org/2000/table" '
			+ 'xmlns:draw="http://openoffice.org/2000/drawing" '
			+ 'xmlns:fo="http://www.w3.org/1999/XSL/Format" '
			+ 'xmlns:xlink="http://www.w3.org/1999/xlink" '
			+ 'xmlns:dc="http://purl.org/dc/elements/1.1/" '
			+ 'xmlns:meta="http://openoffice.org/2000/meta" '
			+ 'xmlns:number="http://openoffice.org/2000/datastyle" '
			+ 'xmlns:svg="http://www.w3.org/2000/svg" '
			+ 'xmlns:chart="http://openoffice.org/2000/chart" '
			+ 'xmlns:dr3d="http://openoffice.org/2000/dr3d" '
			+ 'xmlns:math="http://www.w3.org/1998/Math/MathML" '
			+ 'xmlns:form="http://openoffice.org/2000/form" '
			+ 'xmlns:script="http://openoffice.org/2000/script">',
		)
		.replace("office_color_table", "office:color-table")
		.replace("draw_color", "draw:color")
		.replace("draw_name", "draw:name"),
		encoding="utf-8",
	)


### TXT ###
def openSwatch_TXT(file: str) -> ColourSwatch:
	"""Open a .TXT into a colour swatch"""
	colours = []
	for lineno, line in enumerate(Path(file).read_text(encoding="utf-8").splitlines(False)):
		if line[0] != ";":
			colours.append(
				Colour(
					f"colour{lineno}",
					colour=sRGBColor(
						int(line[0:2], 16), int(line[2:4], 16), int(line[4:6], 16), True
					),
					nameNull=True,
					alpha=int(line[6:8], 16) / 255,
				)
			)
	return getSwatchFromFileName(file, colours)


def saveSwatch_TXT(fileName: str, colourSwatch: ColourSwatch):
	"""Save a colour swatch as .TXT"""
	Path(fileName).write_text(
		"; paint.net Palette File\n; Lines that start with a "
		+ "semicolon are comments\n; Colors are written as 8-digit hexadecimal "
		+ "numbers: aarrggbb\n; For example, this would specify green: FF00FF00\n; "
		+ "The alpha ('aa') value specifies how transparent a color is. FF is "
		+ "fully opaque, 00 is fully transparent.\n; A palette must consist of "
		+ "ninety six (96) colors. If there are less than this, the remaining "
		+ "color\n; slots will be set to white (FFFFFFFF). If there are more, "
		+ "then the remaining colors will be ignored.\n"
		+ "\n".join(
			[
				"".join(colour.getRGB255Hex(True)) + f"{int(colour.alpha * 255):02X}"
				for colour in colourSwatch.colours
			]
		)
		+ "\n",
		encoding="utf-8",
	)


### ACBL ###
def openSwatch_ACBL(file: str) -> ColourSwatch:
	"""Open a .ACBL into a colour swatch"""
	xmlrep = parse(file).getroot()
	cType = xmlrep[1][0].attrib["ColorSpace"]
	colours = []
	for colour in xmlrep[2]:
		cName = colour.attrib["N"]
		cRawColour = [float(col) for col in colour[0].text.split(None)]
		if cType == "CMYK":
			cColour = CMYKColor(*cRawColour)
		elif cType == "RGB":
			cColour = sRGBColor(*cRawColour)
		else:
			cColour = LabColor(*cRawColour)
		colours.append(Colour(cName, cColour))
	return getSwatchFromFileName(file, colours)


def saveSwatch_ACBL(fileName: str, colourSwatch: ColourSwatch):
	"""Save a colour swatch as .ACBL"""
	raise NotImplementedError


### XML ###
def openSwatch_XML(file: str) -> ColourSwatch:
	"""Open a .XML into a colour swatch"""
	xmlrep = parse(file).getroot()
	colours = []
	for colour in xmlrep:
		cName = colour.attrib["NAME"]
		col = colour.attrib["RGB"][1:] if "RGB" in colour.attrib else colour.attrib["CMYK"][1:]
		cColour = (
			sRGBColor(int(col[0:2], 16), int(col[2:4], 16), int(col[4:6], 16), True)
			if "RGB" in colour.attrib
			else CMYKColor(
				int(col[0:2], 16) / 255,
				int(col[2:4], 16) / 255,
				int(col[4:6], 16) / 255,
				int(col[6:8], 16) / 255,
			)
		)
		colours.append(Colour(cName, cColour))
	return ColourSwatch(xmlrep.attrib["Name"], colours)


def saveSwatch_XML(fileName: str, colourSwatch: ColourSwatch):
	"""Save a colour swatch as .XML"""
	root = Element("SCRIBUSCOLORS", {"Name": colourSwatch.name})
	for colour in colourSwatch.colours:
		SubElement(
			root,
			"COLOR",
			{
				"Spot": "0",
				"Register": "0",
				"NAME": colour.name,
				"RGB": "#"
				+ "".join(
					[f"{col:02x}" for col in getWriteOutColour(colour.toRGB().get_value_tuple())]
				),
			},
		)
	Path(fileName).write_text(
		prettify(root, indent=" ", doctype='<?xml version="1.0" encoding="UTF-8"?>'),
		encoding="utf-8",
	)


### PaintShopPro PAL ###
def openSwatch_PSPPAL(file: str) -> ColourSwatch:
	"""Open a PaintShopPro .PAL into a colour swatch"""
	colours = []
	for lineno, line in enumerate(Path(file).read_text(encoding="utf-8").splitlines(False)[3:]):
		colours.append(getColourFromLine(line, lineno))
	return getSwatchFromFileName(file, colours)


def saveSwatch_PSPPAL(fileName: str, colourSwatch: ColourSwatch):
	"""Save a colour swatch as PaintShopPro .PAL"""
	Path(fileName).write_text(
		f"JASC-PAL\n0100\n{len(colourSwatch.colours)}\n"
		+ "\n".join(
			[
				" ".join([f"{col}" for col in getWriteOutColour(colour.toRGB().get_value_tuple())])
				for colour in colourSwatch.colours
			]
		)
		+ "\n"
	)


### CorelDraw PAL ###
def openSwatch_CDPAL(file: str) -> ColourSwatch:
	"""Open a CorelDraw .PAL into a colour swatch"""
	if Path(file).read_text(encoding="utf-8").splitlines(False)[0].strip() == "JASC-PAL":
		return openSwatch_PSPPAL(file)
	colours = []
	for line in Path(file).read_text(encoding="utf-8").splitlines(False):
		parts = split(line)
		colours.append(
			Colour(parts[0], colour=CMYKColor(*[float(col) / 100 for col in parts[1:5]]))
		)
	return getSwatchFromFileName(file, colours)


def saveSwatch_CDPAL(fileName: str, colourSwatch: ColourSwatch):
	"""Save a colour swatch as CorelDraw .PAL"""
	for colour in colourSwatch.colours:
		if colour.nameNull:
			return saveSwatch_PSPPAL(fileName, colourSwatch)
	Path(fileName).write_text(
		"\n".join(
			[
				f'"{colour.name}"  '
				+ "  ".join(
					[
						f"{col:>3}"
						for col in getWriteOutColour(colour.toCMYK().get_value_tuple(), int, 100)
					]
				)
				for colour in colourSwatch.colours
			]
		)
		+ "\n"
	)
	return True


### HPL ###
def openSwatch_HPL(file: str) -> ColourSwatch:
	"""Open a .HPL into a colour swatch"""
	colours = []
	for lineno, line in enumerate(Path(file).read_text(encoding="utf-8").splitlines(False)[3:]):
		colours.append(getColourFromLine(line, lineno))
	return getSwatchFromFileName(file, colours)


def saveSwatch_HPL(fileName: str, colourSwatch: ColourSwatch):
	"""Save a colour swatch as .HPL"""
	Path(fileName).write_text(
		"Palette\nVersion 4.0\n\n"
		+ "\n".join(
			[
				" ".join([f"{col}" for col in getWriteOutColour(colour.toRGB().get_value_tuple())])
				for colour in colourSwatch.colours
			]
		)
		+ "\n",
		encoding="utf-8",
	)


### ASE ###
def openSwatch_ASE(file: str) -> ColourSwatch:
	"""Open an .ase into a list of colour swatches"""
	project = swatch.parse(file)
	swatches = []
	for swtch in project:
		colours = []
		for colour in swtch["swatches"]:
			if colour["data"]["mode"] == "LAB":
				col = LabColor(*colour["data"]["values"])
			elif colour["data"]["mode"] == "RGB":
				col = sRGBColor(*colour["data"]["values"])
			elif colour["data"]["mode"] == "CMYK":
				col = CMYKColor(*colour["data"]["values"])
			else:
				col = sRGBColor(
					colour["data"]["values"][0],
					colour["data"]["values"][0],
					colour["data"]["values"][0],
				)
			colours.append(Colour(colour["name"], colour=col))
		swatches.append(ColourSwatch(swtch["name"], colours))
	return swatches


def saveSwatch_ASE(fileName: str, colourSwatch: ColourSwatch):
	"""Save a colour swatch as .ase"""
	raise NotImplementedError


### IMAGE ###
def openSwatch_PNG(file: str) -> ColourSwatch:
	"""Open a .png into a colour swatch"""
	colours = []
	project = Image.open(file).convert("RGB")
	seen = set()
	pixels = list(project.getdata())
	rawColours = [x for x in pixels if x not in seen and not seen.add(x)]
	for rCol in rawColours:
		colours.append(
			Colour(
				f"colour{len(colours)}",
				colour=sRGBColor(rCol[0], rCol[1], rCol[2], True),
				nameNull=True,
			)
		)
	return getSwatchFromFileName(file, colours)


def openSwatch_IMAGE(file: str) -> ColourSwatch:
	"""open .jpg, .webp"""
	# Colours should be 16x16 px on a canvas with size 256x(16*ceil(colours/16)
	colours = []
	project = Image.open(file).convert("RGB")
	cols, rows = int(project.size[0] / 16), int(project.size[1] / 16)
	for row in range(rows):
		for col in range(cols):
			rCol = project.getpixel((col * 16 + 8, row * 16 + 8))
			colours.append(
				Colour(
					f"colour{len(colours)}",
					colour=sRGBColor(rCol[0], rCol[1], rCol[2], True),
					nameNull=True,
				)
			)
	return getSwatchFromFileName(file, colours)


def saveSwatch_IMAGE(fileName: str, colourSwatch: ColourSwatch):
	"""Save a colour swatch as .png, .jpg, .webp"""
	# Colours should be 16x16 px on a canvas with size 256x(16*ceil(colours/16)
	colours = colourSwatch.colours
	rows = ceil(len(colours) / 16)
	image = Image.new("RGB", (256, 16 * rows), colours[-1].getRGB255())
	draw = ImageDraw.Draw(image)
	index = 0
	for row in range(rows):
		for col in range(16):
			draw.rectangle(
				[col * 16, row * 16, (col + 1) * 16, (row + 1) * 16],
				fill=colours[index].getRGB255(),
				width=0,
			)
			index += 1 if index < len(colours) - 1 else 0
	image.save(fileName)


### SVG ###
def openSwatch_SVG(file: str) -> ColourSwatch:
	"""Open a .svg into a colour swatch"""
	colours = []
	rCols = set()
	# get a list of colours #rgb, #rrggbb. rgb(r,g,b). rgb(r%,g%,b%)
	rColours = findall(
		r"(#......)|(#...)|rgb\((.*?,.*?,.*?)\)", Path(file).read_text(encoding="utf-8")
	)
	for rColour in rColours:
		if len(rColour[0]) > 0:  # #rrggbb
			col = (
				int(rColour[0][1:3], 16) / 255,
				int(rColour[0][3:5], 16) / 255,
				int(rColour[0][5:7], 16) / 255,
			)
		elif len(rColour[1]) > 0:  # #rgb
			col = (
				int(rColour[1][1], 16) / 16,
				int(rColour[1][2], 16) / 16,
				int(rColour[1][3], 16) / 15,
			)
		else:
			r, g, b = rColour[2].split(",")
			r, g, b = r.strip(), g.strip(), b.strip()
			if r[-1] == "%":  # rgb(r%,g%,b%)
				col = (int(r[:-1]) / 100, int(g[:-1]) / 100, int(b[:-1]) / 100)
			else:  # rgb(r,g,b)
				col = (int(r) / 255, int(g) / 255, int(b) / 255)
		if col not in rCols:
			rCols.add(col)
			colours.append(Colour(f"colour{len(colours)}", sRGBColor(col[0], col[1], col[2]), True))
	return getSwatchFromFileName(file, colours)


def saveSwatch_SVG(fileName: str, colourSwatch: ColourSwatch):
	"""Save a colour swatch as .svg"""
	colours = colourSwatch.colours
	rows = ceil(len(colours) / 16)
	colours[-1].toRGB()
	data = [
		(
			'<svg xmlns="http://www.w3.org/2000/svg" height="{0}" '
			+ 'width="256" version="1.1">\n\t<rect style="fill:#{1}" '
			+ 'height="{0}" width="256" x="0" y="0"/>'
		).format(rows * 16, "".join(colours[-1].getRGB255Hex()))
	]
	index = 0
	for row in range(rows):
		for col in range(16):
			colours[index].toRGB()
			data.append(
				'\t<rect style="fill:#{}" height="16" width="16" x="{}" y="{}"/>'.format(
					"".join(colours[index].getRGB255Hex()), col * 16, row * 16
				)
			)
			if index < len(colours) - 2:
				index += 1
			else:
				break
	data.append("</svg>\n")
	Path(fileName).write_text("\n".join(data))



""" hold colourswatch data and provide a series of helper methods such as the
ability to convert to to a pillow palette.
"""

# pylint: disable=too-few-public-methods
# pylint: disable=too-many-arguments
from __future__ import annotations

from colormath.color_conversions import convert_color
from colormath.color_objects import (
	CMYKColor,
	ColorBase,
	HSLColor,
	HSVColor,
	LabColor,
	sRGBColor,
)


class ColourSwatch:
	"""this represents a colour swatch"""

	def __init__(
		self,
		name: str,
		colours: list[Colour] | None = None,
		swatchId: str | None = None,
		description: str | None = None,
		swatchCopyright: str | None = None,
		author: str | None = None,
	):
		self.swatchId = swatchId
		self.name = name
		self.description = description
		self.copyright = swatchCopyright
		self.colours = colours if colours is not None else []
		self.author = author

	def toPILPalette(self) -> list[int]:
		"""Convert the ColourSwatch object to a pil palette

		Usage:
		```python
		image = PIL.Image.new('P', (1, 1))
		image.putpalette(colourSwatch.toPILPalette())
		```
		"""
		pilPalette = []
		for colour in self.colours:
			pilPalette.extend(list(colour.getRGB255()))
		if len(pilPalette) < (256 * 3):
			pilPalette.extend([0] * (256 * 3 - len(pilPalette)))
		else:
			pilPalette = pilPalette[: 256 * 3]
		return pilPalette

	def __repr__(self) -> str:
		"""get a string representation of the object"""
		return '<ColourSwatch "' + self.name + '" colours:' + str(len(self.colours)) + ">"

	def __eq__(self, other: ColourSwatch):
		"""probably not ideal for getting equality - avoid using =="""
		if len(self.colours) != len(other.colours):
			return False
		return True


class Colour:
	"""this represents a single colour within the colour swatch"""

	def __init__(
		self,
		name: str,
		colour: ColorBase | None = None,
		nameNull: bool = False,
		alpha: float = 1.0,
	):
		self.name = name
		self.nameNull = nameNull
		self.colour = colour
		self.alpha = alpha
		self.convertedColour: ColorBase | None = None

	def __repr__(self):
		"""get a string representation of the object"""
		bConverted = (
			self.convertedColour
		)  # do a backup of the convertedColour, we will need to restore
		self.toRGB()
		rString = (
			'<Colour "'
			+ self.name
			+ '" RGB:(hex=#'
			+ "".join(self.convertedColourToHexTuple())
			+ ", dec="
			+ ", ".join([str(col) for col in self.convertedColourToTuple()])
			+ ")>"
		)
		self.convertedColour = bConverted  # and restore
		return rString

	def __eq__(self, other: Colour) -> bool:
		"""equals"""
		return self.toRGB() == other.toRGB()

	def toRGB(self) -> sRGBColor:
		"""convert to rgb and dump a copy in self.convertedColour"""
		self.convertedColour = convert_color(self.colour, sRGBColor)
		return self.convertedColour

	def toCMYK(self) -> CMYKColor:
		"""convert to cmyk and dump a copy in self.convertedColour"""
		self.convertedColour = convert_color(self.colour, CMYKColor)
		return self.convertedColour

	def toHSV(self) -> HSVColor:
		"""convert to hsv and dump a copy in self.convertedColour"""
		self.convertedColour = convert_color(self.colour, HSVColor)
		return self.convertedColour

	def toHSL(self) -> HSLColor:
		"""convert to hsl and dump a copy in self.convertedColour"""
		self.convertedColour = convert_color(self.colour, HSLColor)
		return self.convertedColour

	def toLAB(self) -> LabColor:
		"""convert to lab and dump a copy in self.convertedColour"""
		self.convertedColour = convert_color(self.colour, LabColor)
		return self.convertedColour

	def colorToTuple(self) -> tuple[float, ...]:
		"""get the colour as a tuple. eg. sRGBColor -> (r, g, b)"""
		if not self.colour:
			raise ValueError
		return self.colour.get_value_tuple()

	def convertedColourToTuple(self) -> tuple[float, ...]:
		"""get the previously converted colour as a tuple. eg.
		sRGBColor -> (r, g, b)
		"""
		if not self.convertedColour:
			raise ValueError
		return self.convertedColour.get_value_tuple()

	def convertedColourToHexTuple(self, uppercase: bool = False) -> tuple[str, ...]:
		"""get the previously converted colour as a tuple of hexstrings. eg.
		sRGBColor -> ("ff", "ff", "ff")

		Args:
			uppercase (bool, optional): return hex in uppercase. Defaults to False.

		Returns:
			tuple: tuple of hexstrings
		"""
		return tuple(
			f"{colourPart:02X}" if uppercase else f"{colourPart:02x}"
			for colourPart in self.getRGB255()
		)

	def getRGB255(self) -> tuple[int, ...]:
		"""get the colour as an rgb 255 tuple"""
		self.toRGB()
		return tuple(int(colourPart * 255) for colourPart in self.convertedColourToTuple())

	def getRGB255Hex(self, uppercase: bool = False) -> tuple[str, ...]:
		"""get the colour as an rgb 255 tuple in hex"""
		self.toRGB()
		return self.convertedColourToHexTuple(uppercase)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['colourswatch']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=9.0.0,<10',
 'PyYAML>=6.0,<7',
 'colormath>=3.0.0,<4',
 'defusedxml>=0.7.1,<2',
 'metprint>=2021,<2023',
 'swatch>=0.4.0,<2',
 'tomlkit>=0.8.0,<2']

setup_kwargs = {
    'name': 'colourswatch',
    'version': '2022',
    'description': 'Use this module to read, and write to a number of colour palette file formats',
    'long_description': '[![GitHub top language](https://img.shields.io/github/languages/top/FHPythonUtils/ColourSwatch.svg?style=for-the-badge)](../../)\n[![Repository size](https://img.shields.io/github/repo-size/FHPythonUtils/ColourSwatch.svg?style=for-the-badge)](../../)\n[![Issues](https://img.shields.io/github/issues/FHPythonUtils/ColourSwatch.svg?style=for-the-badge)](../../issues)\n[![License](https://img.shields.io/github/license/FHPythonUtils/ColourSwatch.svg?style=for-the-badge)](/LICENSE.md)\n[![Commit activity](https://img.shields.io/github/commit-activity/m/FHPythonUtils/ColourSwatch.svg?style=for-the-badge)](../../commits/master)\n[![Last commit](https://img.shields.io/github/last-commit/FHPythonUtils/ColourSwatch.svg?style=for-the-badge)](../../commits/master)\n[![PyPI Downloads](https://img.shields.io/pypi/dm/colourswatch.svg?style=for-the-badge)](https://pypistats.org/packages/colourswatch)\n[![PyPI Total Downloads](https://img.shields.io/badge/dynamic/json?style=for-the-badge&label=total%20downloads&query=%24.total_downloads&url=https%3A%2F%2Fapi.pepy.tech%2Fapi%2Fprojects%2Fcolourswatch)](https://pepy.tech/project/colourswatch)\n[![PyPI Version](https://img.shields.io/pypi/v/colourswatch.svg?style=for-the-badge)](https://pypi.org/project/colourswatch)\n\n<!-- omit in TOC -->\n# ColourSwatch\n\n<img src="readme-assets/icons/name.png" alt="Project Icon" width="750">\n\nUse this module to read, and write to a number of colour palette file formats\n\n- [Compatibility](#compatibility)\n\t- [*](#)\n\t- [&](#-1)\n\t- [$](#-2)\n- [Documentation](#documentation)\n- [Install With PIP](#install-with-pip)\n- [Language information](#language-information)\n\t- [Built for](#built-for)\n- [Install Python on Windows](#install-python-on-windows)\n\t- [Chocolatey](#chocolatey)\n\t- [Download](#download)\n- [Install Python on Linux](#install-python-on-linux)\n\t- [Apt](#apt)\n- [How to run](#how-to-run)\n\t- [With VSCode](#with-vscode)\n\t- [From the Terminal](#from-the-terminal)\n- [Download Project](#download-project)\n\t- [Clone](#clone)\n\t\t- [Using The Command Line](#using-the-command-line)\n\t\t- [Using GitHub Desktop](#using-github-desktop)\n\t- [Download Zip File](#download-zip-file)\n- [Community Files](#community-files)\n\t- [Licence](#licence)\n\t- [Changelog](#changelog)\n\t- [Code of Conduct](#code-of-conduct)\n\t- [Contributing](#contributing)\n\t- [Security](#security)\n\t- [Support](#support)\n\t- [Rationale](#rationale)\n\n## Compatibility\n\nBear in mind that the tables below may not be completely accurate. If that is\nthe case, please open an issue and I will fix the tables.\n\n|Application|Extension|Read|Write|\n|---|---|---|---|\n|Gimp|.gpl|✔                |✔                |\n|KOffice|.colors|✔                |✔                |\n|Skencil|.spl|✔                |✔                |\n|sK1|.skp|✔                |✔                |\n|LibreOffice|.soc|✔                |✔                |\n|Paint.net|.txt|✔                |✔                |\n|Adobe Color Book Legacy|.acbl|✔                |❌|\n|Scribus|.xml|✔                |✔                |\n|PaintShopPro|.pal|✔                |✔                |\n|CorelDraw|.pal|✔                |✔                |\n|ColdFusion|.hpl|✔                |✔                |\n|*|.yaml|✔                |✔                |\n|*|.toml|✔                |✔                |\n|*|.json|✔                |✔                |\n|&|.png|✔                |✔                |\n|&|.svg|✔                |✔                |\n|$|.jpg|✔                |✔                |\n|$|.webp|✔                |✔                |\n|Adobe Swatch Exchange|.ase|✔                |❌|\n\n### *\n\nFiles require a name (`name` or `scheme`) and an author is optional (`author`)\n\n### &\n\nSupport for any PNG and any SVG when reading\n\n### $\n\nColours should be 16x16 px on a canvas with size 256x(16*ceil(colours/16)\ne.g.\n\n<img src="test/colours.jpg" width="256">\n\n## Documentation\n\nSee the [Docs](/DOCS/) for more information.\n\n## Install With PIP\n\n```python\npip install colourswatch\n```\n\nHead to https://pypi.org/project/colourswatch/ for more info\n\n## Language information\n\n### Built for\n\nThis program has been written for Python versions 3.7 - 3.10 and has been tested with both 3.7 and\n3.10\n\n## Install Python on Windows\n\n### Chocolatey\n\n```powershell\nchoco install python\n```\n\n### Windows - Python.org\n\nTo install Python, go to https://www.python.org/downloads/windows/ and download the latest\nversion.\n\n## Install Python on Linux\n\n### Apt\n\n```bash\nsudo apt install python3.x\n```\n\n### Dnf\n\n```bash\nsudo dnf install python3.x\n```\n\n## Install Python on MacOS\n\n### Homebrew\n\n```bash\nbrew install python@3.x\n```\n\n### MacOS - Python.org\n\nTo install Python, go to https://www.python.org/downloads/macos/ and download the latest\nversion.\n\n## How to run\n\n### Windows\n\n- Module\n\t`py -3.x -m [module]` or `[module]` (if module installs a script)\n\n- File\n\t`py -3.x [file]` or `./[file]`\n\n### Linux/ MacOS\n\n- Module\n\t`python3.x -m [module]` or `[module]` (if module installs a script)\n\n- File\n\t`python3.x [file]` or `./[file]`\n\n## Download Project\n\n### Clone\n\n#### Using The Command Line\n\n1. Press the Clone or download button in the top right\n2. Copy the URL (link)\n3. Open the command line and change directory to where you wish to\nclone to\n4. Type \'git clone\' followed by URL in step 2\n```bash\ngit clone https://github.com/FHPythonUtils/ColourSwatch\n```\n\nMore information can be found at\nhttps://help.github.com/en/articles/cloning-a-repository\n\n#### Using GitHub Desktop\n\n1. Press the Clone or download button in the top right\n2. Click open in desktop\n3. Choose the path for where you want and click Clone\n\nMore information can be found at\nhttps://help.github.com/en/desktop/contributing-to-projects/cloning-a-repository-from-github-to-github-desktop\n\n### Download Zip File\n\n1. Download this GitHub repository\n2. Extract the zip archive\n3. Copy/ move to the desired location\n\n## Community Files\n\n### Licence\n\nMIT License\nCopyright (c) FredHappyface\n(See the [LICENSE](/LICENSE.md) for more information.)\n\n### Changelog\n\nSee the [Changelog](/CHANGELOG.md) for more information.\n\n### Code of Conduct\n\nOnline communities include people from many backgrounds. The *Project*\ncontributors are committed to providing a friendly, safe and welcoming\nenvironment for all. Please see the\n[Code of Conduct](https://github.com/FHPythonUtils/.github/blob/master/CODE_OF_CONDUCT.md)\n for more information.\n\n### Contributing\n\nContributions are welcome, please see the\n[Contributing Guidelines](https://github.com/FHPythonUtils/.github/blob/master/CONTRIBUTING.md)\nfor more information.\n\n### Security\n\nThank you for improving the security of the project, please see the\n[Security Policy](https://github.com/FHPythonUtils/.github/blob/master/SECURITY.md)\nfor more information.\n\n### Support\n\nThank you for using this project, I hope it is of use to you. Please be aware that\nthose involved with the project often do so for fun along with other commitments\n(such as work, family, etc). Please see the\n[Support Policy](https://github.com/FHPythonUtils/.github/blob/master/SUPPORT.md)\nfor more information.\n\n### Rationale\n\nThe rationale acts as a guide to various processes regarding projects such as\nthe versioning scheme and the programming styles used. Please see the\n[Rationale](https://github.com/FHPythonUtils/.github/blob/master/RATIONALE.md)\nfor more information.\n',
    'author': 'FredHappyface',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/FHPythonUtils/ColourSwatch',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
